@@ -23,7 +23,6 @@
 import time
 import queue
 import os
-import stat
 import errno
 import random
 import re
@@ -32,11 +31,17 @@ from collections import defaultdict
 import threading
 import socket
 import json
+import sys
+import ipaddress
 
+import dns
+import dns.resolver
 import socks
+
 from . import util
+from .util import print_error
 from . import bitcoin
-from .bitcoin import *
+from .bitcoin import COIN
 from . import constants
 from .interface import Connection, Interface
 from . import blockchain
@@ -50,7 +55,6 @@ SERVER_RETRY_INTERVAL = 10
 
 def parse_servers(result):
     """ parse servers list into dict format"""
-    from .version import PROTOCOL_VERSION
     servers = {}
     for item in result:
         host = item[1]
@@ -169,7 +173,7 @@ class Network(util.DaemonThread):
         util.DaemonThread.__init__(self)
         self.config = SimpleConfig(config) if isinstance(config, dict) else config
         self.num_server = 10 if not self.config.get('oneserver') else 0
-        self.blockchains = blockchain.read_blockchains(self.config)
+        self.blockchains = blockchain.read_blockchains(self.config)  # note: needs self.blockchains_lock
         self.print_error("blockchains", self.blockchains.keys())
         self.blockchain_index = config.get('blockchain_index', 0)
         if self.blockchain_index not in self.blockchains.keys():
@@ -185,29 +189,35 @@ class Network(util.DaemonThread):
                 self.default_server = None
         if not self.default_server:
             self.default_server = pick_random_server()
-        self.lock = threading.Lock()
+
+        # locks: if you need to take multiple ones, acquire them in the order they are defined here!
+        self.interface_lock = threading.RLock()            # <- re-entrant
+        self.callback_lock = threading.Lock()
+        self.pending_sends_lock = threading.Lock()
+        self.recent_servers_lock = threading.RLock()       # <- re-entrant
+        self.subscribed_addresses_lock = threading.Lock()
+        self.blockchains_lock = threading.Lock()
+
         self.pending_sends = []
         self.message_id = 0
         self.debug = False
         self.irc_servers = {} # returned by interface (list from irc)
-        self.recent_servers = self.read_recent_servers()
+        self.recent_servers = self.read_recent_servers()  # note: needs self.recent_servers_lock
 
         self.banner = ''
         self.donation_address = ''
         self.relay_fee = None
         # callbacks passed with subscriptions
-        self.subscriptions = defaultdict(list)
-        self.sub_cache = {}
+        self.subscriptions = defaultdict(list)  # note: needs self.callback_lock
+        self.sub_cache = {}                     # note: needs self.interface_lock
         # callbacks set by the GUI
-        self.callbacks = defaultdict(list)
+        self.callbacks = defaultdict(list)      # note: needs self.callback_lock
 
         dir_path = os.path.join( self.config.path, 'certs')
-        if not os.path.exists(dir_path):
-            os.mkdir(dir_path)
-            os.chmod(dir_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
+        util.make_dir(dir_path)
 
         # subscriptions and requests
-        self.subscribed_addresses = set()
+        self.subscribed_addresses = set()  # note: needs self.subscribed_addresses_lock
         self.h2addr = {}
         # Requests from client we've not seen a response to
         self.unanswered_requests = {}
@@ -217,8 +227,8 @@ class Network(util.DaemonThread):
         # kick off the network.  interface is the main server we are currently
         # communicating with.  interfaces is the set of servers we are connecting
         # to or have an ongoing connection with
-        self.interface = None
-        self.interfaces = {}
+        self.interface = None              # note: needs self.interface_lock
+        self.interfaces = {}               # note: needs self.interface_lock
         self.auto_connect = self.config.get('auto_connect', True)
         self.connecting = set()
         self.requested_chunks = set()
@@ -226,19 +236,31 @@ class Network(util.DaemonThread):
         self.start_network(deserialize_server(self.default_server)[2],
                            deserialize_proxy(self.config.get('proxy')))
 
+    def with_interface_lock(func):
+        def func_wrapper(self, *args, **kwargs):
+            with self.interface_lock:
+                return func(self, *args, **kwargs)
+        return func_wrapper
+
+    def with_recent_servers_lock(func):
+        def func_wrapper(self, *args, **kwargs):
+            with self.recent_servers_lock:
+                return func(self, *args, **kwargs)
+        return func_wrapper
+
     def register_callback(self, callback, events):
-        with self.lock:
+        with self.callback_lock:
             for event in events:
                 self.callbacks[event].append(callback)
 
     def unregister_callback(self, callback):
-        with self.lock:
+        with self.callback_lock:
             for callbacks in self.callbacks.values():
                 if callback in callbacks:
                     callbacks.remove(callback)
 
     def trigger_callback(self, event, *args):
-        with self.lock:
+        with self.callback_lock:
             callbacks = self.callbacks[event][:]
         [callback(event, *args) for callback in callbacks]
 
@@ -253,6 +275,7 @@ class Network(util.DaemonThread):
         except:
             return []
 
+    @with_recent_servers_lock
     def save_recent_servers(self):
         if not self.config.path:
             return
@@ -264,6 +287,7 @@ class Network(util.DaemonThread):
         except:
             pass
 
+    @with_interface_lock
     def get_server_height(self):
         return self.interface.tip if self.interface else 0
 
@@ -291,11 +315,15 @@ class Network(util.DaemonThread):
     def is_up_to_date(self):
         return self.unanswered_requests == {}
 
+    @with_interface_lock
     def queue_request(self, method, params, interface=None):
         # If you want to queue a request on any interface it must go
         # through this function so message ids are properly tracked
         if interface is None:
             interface = self.interface
+        if interface is None:
+            self.print_error('warning: dropping request', method, params)
+            return
         message_id = self.message_id
         self.message_id += 1
         if self.debug:
@@ -303,15 +331,14 @@ class Network(util.DaemonThread):
         interface.queue_request(method, params, message_id)
         return message_id
 
+    @with_interface_lock
     def send_subscriptions(self):
+        assert self.interface
         self.print_error('sending subscriptions to', self.interface.server, len(self.unanswered_requests), len(self.subscribed_addresses))
         self.sub_cache.clear()
         # Resend unanswered requests
         requests = self.unanswered_requests.values()
         self.unanswered_requests = {}
-        if self.interface.ping_required():
-            params = [ELECTRUM_VERSION, PROTOCOL_VERSION]
-            self.queue_request('server.version', params, self.interface)
         for request in requests:
             message_id = self.queue_request(request[0], request[1])
             self.unanswered_requests[message_id] = request
@@ -320,8 +347,9 @@ class Network(util.DaemonThread):
         self.queue_request('server.peers.subscribe', [])
         self.request_fee_estimates()
         self.queue_request('blockchain.relayfee', [])
-        for h in list(self.subscribed_addresses):
-            self.queue_request('blockchain.scripthash.subscribe', [h])
+        with self.subscribed_addresses_lock:
+            for h in self.subscribed_addresses:
+                self.queue_request('blockchain.scripthash.subscribe', [h])
 
     def request_fee_estimates(self):
         from .simple_config import FEE_ETA_TARGETS
@@ -361,10 +389,12 @@ class Network(util.DaemonThread):
         if self.is_connected():
             return self.donation_address
 
+    @with_interface_lock
     def get_interfaces(self):
         '''The interfaces that are in connected state'''
         return list(self.interfaces.keys())
 
+    @with_recent_servers_lock
     def get_servers(self):
         out = constants.net.DEFAULT_SERVERS
         if self.irc_servers:
@@ -379,6 +409,7 @@ class Network(util.DaemonThread):
                     out[host] = { protocol:port }
         return out
 
+    @with_interface_lock
     def start_interface(self, server):
         if (not server in self.interfaces and not server in self.connecting):
             if server == self.default_server:
@@ -388,7 +419,8 @@ class Network(util.DaemonThread):
             c = Connection(server, self.socket_queue, self.config.path)
 
     def start_random_interface(self):
-        exclude_set = self.disconnected_servers.union(set(self.interfaces))
+        with self.interface_lock:
+            exclude_set = self.disconnected_servers.union(set(self.interfaces))
         server = pick_random_server(self.get_servers(), self.protocol, exclude_set)
         if server:
             self.start_interface(server)
@@ -418,17 +450,53 @@ class Network(util.DaemonThread):
             socket.getaddrinfo = lambda *args: [(socket.AF_INET, socket.SOCK_STREAM, 6, '', (args[0], args[1]))]
         else:
             socket.socket = socket._socketobject
-            socket.getaddrinfo = socket._getaddrinfo
+            if sys.platform == 'win32':
+                # On Windows, socket.getaddrinfo takes a mutex, and might hold it for up to 10 seconds
+                # when dns-resolving. To speed it up drastically, we resolve dns ourselves, outside that lock.
+                # see #4421
+                socket.getaddrinfo = self._fast_getaddrinfo
+            else:
+                socket.getaddrinfo = socket._getaddrinfo
 
+    @staticmethod
+    def _fast_getaddrinfo(host, *args, **kwargs):
+        def needs_dns_resolving(host2):
+            try:
+                ipaddress.ip_address(host2)
+                return False  # already valid IP
+            except ValueError:
+                pass  # not an IP
+            if str(host) in ('localhost', 'localhost.',):
+                return False
+            return True
+        try:
+            if needs_dns_resolving(host):
+                answers = dns.resolver.query(host)
+                addr = str(answers[0])
+            else:
+                addr = host
+        except dns.exception.DNSException:
+            # dns failed for some reason, e.g. dns.resolver.NXDOMAIN
+            # this is normal. Simply report back failure:
+            raise socket.gaierror(11001, 'getaddrinfo failed')
+        except BaseException as e:
+            # Possibly internal error in dnspython :( see #4483
+            # Fall back to original socket.getaddrinfo to resolve dns.
+            print_error('dnspython failed to resolve dns with error:', e)
+            addr = host
+        return socket._getaddrinfo(addr, *args, **kwargs)
+
+    @with_interface_lock
     def start_network(self, protocol, proxy):
         assert not self.interface and not self.interfaces
         assert not self.connecting and self.socket_queue.empty()
         self.print_error('starting network')
-        self.disconnected_servers = set([])
+        self.disconnected_servers = set([])  # note: needs self.interface_lock
         self.protocol = protocol
         self.set_proxy(proxy)
         self.start_interfaces()
 
+    @with_interface_lock
     def stop_network(self):
         self.print_error("stopping network")
         for interface in list(self.interfaces.values()):
@@ -478,6 +546,7 @@ class Network(util.DaemonThread):
         if servers:
             self.switch_to_interface(random.choice(servers))
 
+    @with_interface_lock
     def switch_lagging_interface(self):
         '''If auto_connect and lagging, switch interface'''
         if self.server_is_lagging() and self.auto_connect:
@@ -488,6 +557,7 @@ class Network(util.DaemonThread):
                 choice = random.choice(filtered)
                 self.switch_to_interface(choice)
 
+    @with_interface_lock
     def switch_to_interface(self, server):
         '''Switch to server as our interface.  If no connection exists nor
         being opened, start a thread to connect.  The actual switch will
@@ -508,7 +578,9 @@ class Network(util.DaemonThread):
             self.send_subscriptions()
             self.set_status('connected')
             self.notify('updated')
+            self.notify('interfaces')
 
+    @with_interface_lock
     def close_interface(self, interface):
         if interface:
             if interface.server in self.interfaces:
@@ -517,6 +589,7 @@ class Network(util.DaemonThread):
                 self.interface = None
             interface.close()
 
+    @with_recent_servers_lock
     def add_recent_server(self, server):
         # list is ordered
         if server in self.recent_servers:
@@ -539,6 +612,10 @@ class Network(util.DaemonThread):
         elif method == 'blockchain.headers.subscribe':
             if error is None:
                 self.on_notify_header(interface, result)
+            else:
+                # no point in keeping this connection without headers sub
+                self.connection_down(interface.server)
+                return
         elif method == 'server.peers.subscribe':
             if error is None:
                 self.irc_servers = parse_servers(result)
@@ -566,15 +643,16 @@ class Network(util.DaemonThread):
             if error is None:
                 self.relay_fee = int(result * COIN) if result is not None else None
                 self.print_error("relayfee", self.relay_fee)
-        elif method == 'blockchain.block.get_chunk':
-            self.on_get_chunk(interface, response)
+        elif method == 'blockchain.block.headers':
+            self.on_block_headers(interface, response)
         elif method == 'blockchain.block.get_header':
             self.on_get_header(interface, response)
 
         for callback in callbacks:
             callback(response)
 
-    def get_index(self, method, params):
+    @classmethod
+    def get_index(cls, method, params):
         """ hashable index for subscriptions and cache"""
         return str(method) + (':' + str(params[0]) if params else '')
 
@@ -589,12 +667,15 @@ class Network(util.DaemonThread):
                 # and are placed in the unanswered_requests dictionary
                 client_req = self.unanswered_requests.pop(message_id, None)
                 if client_req:
-                    assert interface == self.interface
+                    if interface != self.interface:
+                        # we probably changed the current interface
+                        # in the meantime; drop this.
+                        return
                     callbacks = [client_req[2]]
                 else:
                     # fixme: will only work for subscriptions
                     k = self.get_index(method, params)
-                    callbacks = self.subscriptions.get(k, [])
+                    callbacks = list(self.subscriptions.get(k, []))
 
                 # Copy the request method and params to the response
                 response['method'] = method
@@ -602,7 +683,8 @@ class Network(util.DaemonThread):
                 # Only once we've received a response to an addr subscription
                 # add it to the list; avoids double-sends on reconnection
                 if method == 'blockchain.scripthash.subscribe':
-                    self.subscribed_addresses.add(params[0])
+                    with self.subscribed_addresses_lock:
+                        self.subscribed_addresses.add(params[0])
             else:
                 if not response:  # Closed remotely / misbehaving
                     self.connection_down(interface.server)
@@ -617,51 +699,29 @@ class Network(util.DaemonThread):
                 elif method == 'blockchain.scripthash.subscribe':
                     response['params'] = [params[0]]  # addr
                     response['result'] = params[1]
-                callbacks = self.subscriptions.get(k, [])
+                callbacks = list(self.subscriptions.get(k, []))
 
             # update cache if it's a subscription
             if method.endswith('.subscribe'):
-                self.sub_cache[k] = response
+                with self.interface_lock:
+                    self.sub_cache[k] = response
             # Response is now in canonical form
             self.process_response(interface, response, callbacks)
 
-    def addr_to_scripthash(self, addr):
-        h = bitcoin.address_to_scripthash(addr)
-        if h not in self.h2addr:
-            self.h2addr[h] = addr
-        return h
-
-    def overload_cb(self, callback):
-        def cb2(x):
-            x2 = x.copy()
-            p = x2.pop('params')
-            addr = self.h2addr[p[0]]
-            x2['params'] = [addr]
-            callback(x2)
-        return cb2
-
-    def subscribe_to_addresses(self, addresses, callback):
-        hashes = [self.addr_to_scripthash(addr) for addr in addresses]
-        msgs = [('blockchain.scripthash.subscribe', [x]) for x in hashes]
-        self.send(msgs, self.overload_cb(callback))
-
-    def request_address_history(self, address, callback):
-        h = self.addr_to_scripthash(address)
-        self.send([('blockchain.scripthash.get_history', [h])], self.overload_cb(callback))
-
     def send(self, messages, callback):
         '''Messages is a list of (method, params) tuples'''
         messages = list(messages)
-        with self.lock:
+        with self.pending_sends_lock:
             self.pending_sends.append((messages, callback))
 
+    @with_interface_lock
     def process_pending_sends(self):
         # Requests needs connectivity.  If we don't have an interface,
         # we cannot process them.
         if not self.interface:
             return
 
-        with self.lock:
+        with self.pending_sends_lock:
             sends = self.pending_sends
             self.pending_sends = []
 
@@ -671,12 +731,14 @@ class Network(util.DaemonThread):
                 if method.endswith('.subscribe'):
                     k = self.get_index(method, params)
                     # add callback to list
-                    l = self.subscriptions.get(k, [])
+                    l = list(self.subscriptions.get(k, []))
                     if callback not in l:
                         l.append(callback)
-                    self.subscriptions[k] = l
+                    with self.callback_lock:
+                        self.subscriptions[k] = l
                     # check cached response for subscriptions
                     r = self.sub_cache.get(k)
+
                 if r is not None:
                     self.print_error("cache hit", k)
                     callback(r)
@@ -689,11 +751,12 @@ class Network(util.DaemonThread):
         # Note: we can't unsubscribe from the server, so if we receive
         # subsequent notifications process_response() will emit a harmless
         # "received unexpected notification" warning
-        with self.lock:
+        with self.callback_lock:
             for v in self.subscriptions.values():
                 if callback in v:
                     v.remove(callback)
 
+    @with_interface_lock
     def connection_down(self, server):
         '''A connection to server either went down, or was never made.
         We distinguish by whether it is in self.interfaces.'''
@@ -703,9 +766,10 @@ class Network(util.DaemonThread):
         if server in self.interfaces:
             self.close_interface(self.interfaces[server])
             self.notify('interfaces')
-        for b in self.blockchains.values():
-            if b.catch_up == server:
-                b.catch_up = None
+        with self.blockchains_lock:
+            for b in self.blockchains.values():
+                if b.catch_up == server:
+                    b.catch_up = None
 
     def new_interface(self, server, socket):
         # todo: get tip first, then decide which checkpoint to use.
@@ -716,8 +780,12 @@ class Network(util.DaemonThread):
         interface.tip = 0
         interface.mode = 'default'
         interface.request = None
-        self.interfaces[server] = interface
-        self.queue_request('blockchain.headers.subscribe', [], interface)
+        with self.interface_lock:
+            self.interfaces[server] = interface
+        # server.version should be the first message
+        params = [ELECTRUM_VERSION, PROTOCOL_VERSION]
+        self.queue_request('server.version', params, interface)
+        self.queue_request('blockchain.headers.subscribe', [True], interface)
         if server == self.default_server:
             self.switch_to_interface(server)
         #self.notify('interfaces')
@@ -736,46 +804,51 @@ class Network(util.DaemonThread):
 
         # Send pings and shut down stale interfaces
         # must use copy of values
-        for interface in list(self.interfaces.values()):
+        with self.interface_lock:
+            interfaces = list(self.interfaces.values())
+        for interface in interfaces:
             if interface.has_timed_out():
                 self.connection_down(interface.server)
             elif interface.ping_required():
-                params = [ELECTRUM_VERSION, PROTOCOL_VERSION]
-                self.queue_request('server.version', params, interface)
+                self.queue_request('server.ping', [], interface)
 
         now = time.time()
         # nodes
-        if len(self.interfaces) + len(self.connecting) < self.num_server:
-            self.start_random_interface()
-            if now - self.nodes_retry_time > NODES_RETRY_INTERVAL:
-                self.print_error('network: retrying connections')
-                self.disconnected_servers = set([])
-                self.nodes_retry_time = now
+        with self.interface_lock:
+            if len(self.interfaces) + len(self.connecting) < self.num_server:
+                self.start_random_interface()
+                if now - self.nodes_retry_time > NODES_RETRY_INTERVAL:
+                    self.print_error('network: retrying connections')
+                    self.disconnected_servers = set([])
+                    self.nodes_retry_time = now
 
         # main interface
-        if not self.is_connected():
-            if self.auto_connect:
-                if not self.is_connecting():
-                    self.switch_to_random_interface()
-            else:
-                if self.default_server in self.disconnected_servers:
-                    if now - self.server_retry_time > SERVER_RETRY_INTERVAL:
-                        self.disconnected_servers.remove(self.default_server)
-                        self.server_retry_time = now
+        with self.interface_lock:
+            if not self.is_connected():
+                if self.auto_connect:
+                    if not self.is_connecting():
+                        self.switch_to_random_interface()
                 else:
-                    self.switch_to_interface(self.default_server)
-        else:
-            if self.config.is_fee_estimates_update_required():
-                self.request_fee_estimates()
+                    if self.default_server in self.disconnected_servers:
+                        if now - self.server_retry_time > SERVER_RETRY_INTERVAL:
+                            self.disconnected_servers.remove(self.default_server)
+                            self.server_retry_time = now
+                    else:
+                        self.switch_to_interface(self.default_server)
+            else:
+                if self.config.is_fee_estimates_update_required():
+                    self.request_fee_estimates()
 
     def request_chunk(self, interface, index):
         if index in self.requested_chunks:
             return
         interface.print_error("requesting chunk %d" % index)
         self.requested_chunks.add(index)
-        self.queue_request('blockchain.block.get_chunk', [index], interface)
+        height = index * 2016
+        self.queue_request('blockchain.block.headers', [height, 2016],
+                           interface)
 
-    def on_get_chunk(self, interface, response):
+    def on_block_headers(self, interface, response):
         '''Handle receiving a chunk of block headers'''
         error = response.get('error')
         result = response.get('result')
@@ -784,15 +857,17 @@ class Network(util.DaemonThread):
         if result is None or params is None or error is not None:
             interface.print_error(error or 'bad response')
             return
-        index = params[0]
         # Ignore unsolicited chunks
-        if index not in self.requested_chunks:
+        height = params[0]
+        index = height // 2016
+        if index * 2016 != height or index not in self.requested_chunks:
             interface.print_error("received chunk %d (unsolicited)" % index)
             return
         else:
             interface.print_error("received chunk %d" % index)
         self.requested_chunks.remove(index)
-        connect = blockchain.connect_chunk(index, result)
+        hexdata = result['hex']
+        connect = blockchain.connect_chunk(index, hexdata)
         if not connect:
             self.connection_down(interface.server)
             return
@@ -805,12 +880,6 @@ class Network(util.DaemonThread):
             blockchain.catch_up = None
         self.notify('updated')
 
-    def request_header(self, interface, height):
-        #interface.print_error("requesting header %d" % height)
-        self.queue_request('blockchain.block.get_header', [height], interface)
-        interface.request = height
-        interface.req_time = time.time()
-
     def on_get_header(self, interface, response):
         '''Handle receiving a single block header'''
         header = response.get('result')
@@ -848,6 +917,9 @@ class Network(util.DaemonThread):
                     interface.bad_header = header
                     delta = interface.tip - height
                     next_height = max(self.max_checkpoint(), interface.tip - 2 * delta)
+                    if height == next_height:
+                        self.connection_down(interface.server)
+                        next_height = None
 
         elif interface.mode == 'binary':
             if chain:
@@ -874,7 +946,7 @@ class Network(util.DaemonThread):
                         next_height = None
                     else:
                         interface.print_error('checkpoint conflicts with existing fork', branch.path())
-                        branch.write('', 0)
+                        branch.write(b'', 0)
                         branch.save_header(interface.bad_header)
                         interface.mode = 'catch_up'
                         interface.blockchain = branch
@@ -886,7 +958,8 @@ class Network(util.DaemonThread):
                     if bh > interface.good:
                         if not interface.blockchain.check_header(interface.bad_header):
                             b = interface.blockchain.fork(interface.bad_header)
-                            self.blockchains[interface.bad] = b
+                            with self.blockchains_lock:
+                                self.blockchains[interface.bad] = b
                             interface.blockchain = b
                             interface.print_error("new chain", b.checkpoint)
                             interface.mode = 'catch_up'
@@ -925,7 +998,7 @@ class Network(util.DaemonThread):
         else:
             raise Exception(interface.mode)
         # If not finished, get the next header
-        if next_height:
+        if next_height is not None:
             if interface.mode == 'catch_up' and interface.tip > next_height + 50:
                 self.request_chunk(interface, next_height // 2016)
             else:
@@ -938,7 +1011,9 @@ class Network(util.DaemonThread):
         self.notify('interfaces')
 
     def maintain_requests(self):
-        for interface in list(self.interfaces.values()):
+        with self.interface_lock:
+            interfaces = list(self.interfaces.values())
+        for interface in interfaces:
             if interface.request and time.time() - interface.request_time > 20:
                 interface.print_error("blockchain request timed out")
                 self.connection_down(interface.server)
@@ -950,14 +1025,14 @@ class Network(util.DaemonThread):
         if not self.interfaces:
             time.sleep(0.1)
             return
-        rin = [i for i in self.interfaces.values()]
-        win = [i for i in self.interfaces.values() if i.num_requests()]
+        with self.interface_lock:
+            interfaces = list(self.interfaces.values())
+        rin = [i for i in interfaces]
+        win = [i for i in interfaces if i.num_requests()]
         try:
             rout, wout, xout = select.select(rin, win, [], 0.1)
         except socket.error as e:
-            # TODO: py3, get code from e
-            code = None
-            if code == errno.EINTR:
+            if e.errno == errno.EINTR:
                 return
             raise
         assert not xout
@@ -989,10 +1064,9 @@ class Network(util.DaemonThread):
         self.stop_network()
         self.on_stop()
 
-    def on_notify_header(self, interface, header):
-        height = header.get('block_height')
-        if not height:
-            return
+    def on_notify_header(self, interface, header_dict):
+        header_hex, height = header_dict['hex'], header_dict['height']
+        header = blockchain.deserialize_header(util.bfh(header_hex), height)
         if height < self.max_checkpoint():
             self.connection_down(interface.server)
             return
@@ -1015,7 +1089,8 @@ class Network(util.DaemonThread):
             self.notify('updated')
             self.notify('interfaces')
             return
-        tip = max([x.height() for x in self.blockchains.values()])
+        with self.blockchains_lock:
+            tip = max([x.height() for x in self.blockchains.values()])
         if tip >=0:
             interface.mode = 'backward'
             interface.bad = height
@@ -1027,19 +1102,24 @@ class Network(util.DaemonThread):
                 chain.catch_up = interface
                 interface.mode = 'catch_up'
                 interface.blockchain = chain
-                self.print_error("switching to catchup mode", tip,  self.blockchains)
+                with self.blockchains_lock:
+                    self.print_error("switching to catchup mode", tip,  self.blockchains)
                 self.request_header(interface, 0)
             else:
                 self.print_error("chain already catching up with", chain.catch_up.server)
 
+    @with_interface_lock
     def blockchain(self):
         if self.interface and self.interface.blockchain is not None:
             self.blockchain_index = self.interface.blockchain.checkpoint
         return self.blockchains[self.blockchain_index]
 
+    @with_interface_lock
     def get_blockchains(self):
         out = {}
-        for k, b in self.blockchains.items():
+        with self.blockchains_lock:
+            blockchain_items = list(self.blockchains.items())
+        for k, b in blockchain_items:
             r = list(filter(lambda i: i.blockchain==b, list(self.interfaces.values())))
             if r:
                 out[k] = r
@@ -1050,48 +1130,159 @@ class Network(util.DaemonThread):
         if blockchain:
             self.blockchain_index = index
             self.config.set_key('blockchain_index', index)
-            for i in self.interfaces.values():
+            with self.interface_lock:
+                interfaces = list(self.interfaces.values())
+            for i in interfaces:
                 if i.blockchain == blockchain:
                     self.switch_to_interface(i.server)
                     break
         else:
             raise Exception('blockchain not found', index)
 
-        if self.interface:
-            server = self.interface.server
-            host, port, protocol, proxy, auto_connect = self.get_parameters()
-            host, port, protocol = server.split(':')
-            self.set_parameters(host, port, protocol, proxy, auto_connect)
+        with self.interface_lock:
+            if self.interface:
+                server = self.interface.server
+                host, port, protocol, proxy, auto_connect = self.get_parameters()
+                host, port, protocol = server.split(':')
+                self.set_parameters(host, port, protocol, proxy, auto_connect)
 
     def get_local_height(self):
         return self.blockchain().height()
 
-    def synchronous_get(self, request, timeout=30):
+    @staticmethod
+    def __wait_for(it):
+        """Wait for the result of calling lambda `it`."""
         q = queue.Queue()
-        self.send([request], q.put)
+        it(q.put)
         try:
-            r = q.get(True, timeout)
+            result = q.get(block=True, timeout=30)
         except queue.Empty:
             raise util.TimeoutException(_('Server did not answer'))
-        if r.get('error'):
-            raise Exception(r.get('error'))
-        return r.get('result')
 
-    def broadcast(self, tx, timeout=30):
-        tx_hash = tx.txid()
+        if result.get('error'):
+            raise Exception(result.get('error'))
+
+        return result.get('result')
+
+    @staticmethod
+    def __with_default_synchronous_callback(invocation, callback):
+        """ Use this method if you want to make the network request
+        synchronous. """
+        if not callback:
+            return Network.__wait_for(invocation)
+
+        invocation(callback)
+
+    def request_header(self, interface, height):
+        self.queue_request('blockchain.block.get_header', [height], interface)
+        interface.request = height
+        interface.req_time = time.time()
+
+    def map_scripthash_to_address(self, callback):
+        def cb2(x):
+            x2 = x.copy()
+            p = x2.pop('params')
+            addr = self.h2addr[p[0]]
+            x2['params'] = [addr]
+            callback(x2)
+        return cb2
+
+    def subscribe_to_addresses(self, addresses, callback):
+        hash2address = {
+            bitcoin.address_to_scripthash(address): address
+            for address in addresses}
+        self.h2addr.update(hash2address)
+        msgs = [
+            ('blockchain.scripthash.subscribe', [x])
+            for x in hash2address.keys()]
+        self.send(msgs, self.map_scripthash_to_address(callback))
+
+    def request_address_history(self, address, callback):
+        h = bitcoin.address_to_scripthash(address)
+        self.h2addr.update({h: address})
+        self.send([('blockchain.scripthash.get_history', [h])], self.map_scripthash_to_address(callback))
+
+    # NOTE this method handles exceptions and a special edge case, counter to
+    # what the other ElectrumX methods do. This is unexpected.
+    def broadcast_transaction(self, transaction, callback=None):
+        command = 'blockchain.transaction.broadcast'
+        invocation = lambda c: self.send([(command, [str(transaction)])], c)
+
+        if callback:
+            invocation(callback)
+            return
+
         try:
-            out = self.synchronous_get(('blockchain.transaction.broadcast', [str(tx)]), timeout)
+            out = Network.__wait_for(invocation)
         except BaseException as e:
             return False, "error: " + str(e)
-        if out != tx_hash:
+
+        if out != transaction.txid():
             return False, "error: " + out
+
         return True, out
 
+    def get_history_for_scripthash(self, hash, callback=None):
+        command = 'blockchain.scripthash.get_history'
+        invocation = lambda c: self.send([(command, [hash])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def subscribe_to_headers(self, callback=None):
+        command = 'blockchain.headers.subscribe'
+        invocation = lambda c: self.send([(command, [True])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def subscribe_to_address(self, address, callback=None):
+        command = 'blockchain.address.subscribe'
+        invocation = lambda c: self.send([(command, [address])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def get_merkle_for_transaction(self, tx_hash, tx_height, callback=None):
+        command = 'blockchain.transaction.get_merkle'
+        invocation = lambda c: self.send([(command, [tx_hash, tx_height])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def subscribe_to_scripthash(self, scripthash, callback=None):
+        command = 'blockchain.scripthash.subscribe'
+        invocation = lambda c: self.send([(command, [scripthash])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def get_transaction(self, transaction_hash, callback=None):
+        command = 'blockchain.transaction.get'
+        invocation = lambda c: self.send([(command, [transaction_hash])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def get_transactions(self, transaction_hashes, callback=None):
+        command = 'blockchain.transaction.get'
+        messages = [(command, [tx_hash]) for tx_hash in transaction_hashes]
+        invocation = lambda c: self.send(messages, c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def listunspent_for_scripthash(self, scripthash, callback=None):
+        command = 'blockchain.scripthash.listunspent'
+        invocation = lambda c: self.send([(command, [scripthash])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
+    def get_balance_for_scripthash(self, scripthash, callback=None):
+        command = 'blockchain.scripthash.get_balance'
+        invocation = lambda c: self.send([(command, [scripthash])], c)
+
+        return Network.__with_default_synchronous_callback(invocation, callback)
+
     def export_checkpoints(self, path):
         # run manually from the console to generate checkpoints
         cp = self.blockchain().get_checkpoints()
         with open(path, 'w', encoding='utf-8') as f:
             f.write(json.dumps(cp, indent=4))
 
-    def max_checkpoint(self):
+    @classmethod
+    def max_checkpoint(cls):
         return max(0, len(constants.net.CHECKPOINTS) * 2016 - 1)
