@@ -1,4 +1,4 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python
 #
 # Electrum - lightweight Bitcoin client
 # Copyright (C) 2012 thomasv@gitorious
@@ -22,35 +22,40 @@
 # ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
+
 import socket
 
-from PyQt4.QtGui import *
-from PyQt4.QtCore import *
-import PyQt4.QtCore as QtCore
+from PyQt5.QtGui import *
+from PyQt5.QtCore import *
+from PyQt5.QtWidgets import *
+import PyQt5.QtCore as QtCore
 
 from electrum.i18n import _
-from electrum.network import DEFAULT_PORTS
+from electrum import constants
+from electrum.util import print_error
 from electrum.network import serialize_server, deserialize_server
 
-from util import *
+from .util import *
 
 protocol_names = ['TCP', 'SSL']
 protocol_letters = 'ts'
 
 class NetworkDialog(QDialog):
-    def __init__(self, network, config):
+    def __init__(self, network, config, network_updated_signal_obj):
         QDialog.__init__(self)
         self.setWindowTitle(_('Network'))
         self.setMinimumSize(500, 20)
         self.nlayout = NetworkChoiceLayout(network, config)
+        self.network_updated_signal_obj = network_updated_signal_obj
         vbox = QVBoxLayout(self)
         vbox.addLayout(self.nlayout.layout())
         vbox.addLayout(Buttons(CloseButton(self)))
-        self.connect(self, QtCore.SIGNAL('updated'), self.on_update)
+        self.network_updated_signal_obj.network_updated_signal.connect(
+            self.on_update)
         network.register_callback(self.on_network, ['updated', 'interfaces'])
 
     def on_network(self, event, *args):
-        self.emit(QtCore.SIGNAL('updated'), event, *args)
+        self.network_updated_signal_obj.network_updated_signal.emit(event, args)
 
     def on_update(self):
         self.nlayout.update()
@@ -62,7 +67,7 @@ class NodesListWidget(QTreeWidget):
     def __init__(self, parent):
         QTreeWidget.__init__(self)
         self.parent = parent
-        self.setHeaderLabels([_('Node'), _('Height')])
+        self.setHeaderLabels([_('Connected node'), _('Height')])
         self.setContextMenuPolicy(Qt.CustomContextMenu)
         self.customContextMenuRequested.connect(self.create_menu)
 
@@ -70,13 +75,13 @@ class NodesListWidget(QTreeWidget):
         item = self.currentItem()
         if not item:
             return
-        is_server = not bool(item.data(0, Qt.UserRole).toInt()[0])
+        is_server = not bool(item.data(0, Qt.UserRole))
         menu = QMenu()
         if is_server:
-            server = unicode(item.data(1, Qt.UserRole).toString())
+            server = item.data(1, Qt.UserRole)
             menu.addAction(_("Use as server"), lambda: self.parent.follow_server(server))
         else:
-            index = item.data(1, Qt.UserRole).toInt()[0]
+            index = item.data(1, Qt.UserRole)
             menu.addAction(_("Follow this branch"), lambda: self.parent.follow_branch(index))
         menu.exec_(self.viewport().mapToGlobal(position))
 
@@ -90,7 +95,7 @@ class NodesListWidget(QTreeWidget):
         # on 'enter' we show the menu
         pt = self.visualItemRect(item).bottomLeft()
         pt.setX(50)
-        self.emit(SIGNAL('customContextMenuRequested(const QPoint&)'), pt)
+        self.customContextMenuRequested.emit(pt)
 
     def update(self, network):
         self.clear()
@@ -118,8 +123,8 @@ class NodesListWidget(QTreeWidget):
 
         h = self.header()
         h.setStretchLastSection(False)
-        h.setResizeMode(0, QHeaderView.Stretch)
-        h.setResizeMode(1, QHeaderView.ResizeToContents)
+        h.setSectionResizeMode(0, QHeaderView.Stretch)
+        h.setSectionResizeMode(1, QHeaderView.ResizeToContents)
 
 
 class ServerListWidget(QTreeWidget):
@@ -136,12 +141,12 @@ class ServerListWidget(QTreeWidget):
         if not item:
             return
         menu = QMenu()
-        server = unicode(item.data(1, Qt.UserRole).toString())
+        server = item.data(1, Qt.UserRole)
         menu.addAction(_("Use as server"), lambda: self.set_server(server))
         menu.exec_(self.viewport().mapToGlobal(position))
 
     def set_server(self, s):
-        host, port, protocol = s.split(':')
+        host, port, protocol = deserialize_server(s)
         self.parent.server_host.setText(host)
         self.parent.server_port.setText(port)
         self.parent.set_server()
@@ -156,7 +161,7 @@ class ServerListWidget(QTreeWidget):
         # on 'enter' we show the menu
         pt = self.visualItemRect(item).bottomLeft()
         pt.setX(50)
-        self.emit(SIGNAL('customContextMenuRequested(const QPoint&)'), pt)
+        self.customContextMenuRequested.emit(pt)
 
     def update(self, servers, protocol, use_tor):
         self.clear()
@@ -166,14 +171,14 @@ class ServerListWidget(QTreeWidget):
             port = d.get(protocol)
             if port:
                 x = QTreeWidgetItem([_host, port])
-                server = _host+':'+port+':'+protocol
+                server = serialize_server(_host, port, protocol)
                 x.setData(1, Qt.UserRole, server)
                 self.addTopLevelItem(x)
 
         h = self.header()
         h.setStretchLastSection(False)
-        h.setResizeMode(0, QHeaderView.Stretch)
-        h.setResizeMode(1, QHeaderView.ResizeToContents)
+        h.setSectionResizeMode(0, QHeaderView.Stretch)
+        h.setSectionResizeMode(1, QHeaderView.ResizeToContents)
 
 
 class NetworkChoiceLayout(object):
@@ -189,8 +194,8 @@ class NetworkChoiceLayout(object):
         proxy_tab = QWidget()
         blockchain_tab = QWidget()
         tabs.addTab(blockchain_tab, _('Overview'))
-        tabs.addTab(proxy_tab, _('Proxy'))
         tabs.addTab(server_tab, _('Server'))
+        tabs.addTab(proxy_tab, _('Proxy'))
 
         # server tab
         grid = QGridLayout(server_tab)
@@ -200,16 +205,21 @@ class NetworkChoiceLayout(object):
         self.server_host.setFixedWidth(200)
         self.server_port = QLineEdit()
         self.server_port.setFixedWidth(60)
-        self.ssl_cb = QCheckBox(_('Use SSL'))
         self.autoconnect_cb = QCheckBox(_('Select server automatically'))
         self.autoconnect_cb.setEnabled(self.config.is_modifiable('auto_connect'))
 
         self.server_host.editingFinished.connect(self.set_server)
         self.server_port.editingFinished.connect(self.set_server)
-        self.ssl_cb.clicked.connect(self.change_protocol)
         self.autoconnect_cb.clicked.connect(self.set_server)
         self.autoconnect_cb.clicked.connect(self.update)
 
+        msg = ' '.join([
+            _("If auto-connect is enabled, Electrum will always use a server that is on the longest blockchain."),
+            _("If it is disabled, you have to choose a server you want to use. Electrum will warn you if your server is lagging.")
+        ])
+        grid.addWidget(self.autoconnect_cb, 0, 0, 1, 3)
+        grid.addWidget(HelpButton(msg), 0, 4)
+
         grid.addWidget(QLabel(_('Server') + ':'), 1, 0)
         grid.addWidget(self.server_host, 1, 1, 1, 2)
         grid.addWidget(self.server_port, 1, 3)
@@ -224,8 +234,12 @@ class NetworkChoiceLayout(object):
         grid.setSpacing(8)
 
         # proxy setting
+        self.proxy_cb = QCheckBox(_('Use proxy'))
+        self.proxy_cb.clicked.connect(self.check_disable_proxy)
+        self.proxy_cb.clicked.connect(self.set_proxy)
+
         self.proxy_mode = QComboBox()
-        self.proxy_mode.addItems(['NONE', 'SOCKS4', 'SOCKS5', 'HTTP'])
+        self.proxy_mode.addItems(['SOCKS4', 'SOCKS5', 'HTTP'])
         self.proxy_host = QLineEdit()
         self.proxy_host.setFixedWidth(200)
         self.proxy_port = QLineEdit()
@@ -243,28 +257,26 @@ class NetworkChoiceLayout(object):
         self.proxy_user.editingFinished.connect(self.set_proxy)
         self.proxy_password.editingFinished.connect(self.set_proxy)
 
-        self.check_disable_proxy()
-        self.proxy_mode.connect(self.proxy_mode, SIGNAL('currentIndexChanged(int)'), self.check_disable_proxy)
-
-        self.proxy_mode.connect(self.proxy_mode, SIGNAL('currentIndexChanged(int)'), self.proxy_settings_changed)
-        self.proxy_host.connect(self.proxy_host, SIGNAL('textEdited(QString)'), self.proxy_settings_changed)
-        self.proxy_port.connect(self.proxy_port, SIGNAL('textEdited(QString)'), self.proxy_settings_changed)
-        self.proxy_user.connect(self.proxy_user, SIGNAL('textEdited(QString)'), self.proxy_settings_changed)
-        self.proxy_password.connect(self.proxy_password, SIGNAL('textEdited(QString)'), self.proxy_settings_changed)
+        self.proxy_mode.currentIndexChanged.connect(self.proxy_settings_changed)
+        self.proxy_host.textEdited.connect(self.proxy_settings_changed)
+        self.proxy_port.textEdited.connect(self.proxy_settings_changed)
+        self.proxy_user.textEdited.connect(self.proxy_settings_changed)
+        self.proxy_password.textEdited.connect(self.proxy_settings_changed)
 
         self.tor_cb = QCheckBox(_("Use Tor Proxy"))
         self.tor_cb.setIcon(QIcon(":icons/tor_logo.png"))
         self.tor_cb.hide()
         self.tor_cb.clicked.connect(self.use_tor_proxy)
 
-        grid.addWidget(self.ssl_cb, 0, 0, 1, 3)
         grid.addWidget(self.tor_cb, 1, 0, 1, 3)
+        grid.addWidget(self.proxy_cb, 2, 0, 1, 3)
+        grid.addWidget(HelpButton(_('Proxy settings apply to all connections: with Electrum servers, but also with third-party services.')), 2, 4)
         grid.addWidget(self.proxy_mode, 4, 1)
         grid.addWidget(self.proxy_host, 4, 2)
         grid.addWidget(self.proxy_port, 4, 3)
         grid.addWidget(self.proxy_user, 5, 2)
         grid.addWidget(self.proxy_password, 5, 3)
-        grid.setRowStretch(6, 1)
+        grid.setRowStretch(7, 1)
 
         # Blockchain Tab
         grid = QGridLayout(blockchain_tab)
@@ -292,12 +304,6 @@ class NetworkChoiceLayout(object):
         self.split_label = QLabel('')
         grid.addWidget(self.split_label, 3, 0, 1, 3)
 
-        msg = ' '.join([
-            _("If auto-connect is enabled, Electrum will always use a server that is on the longest blockchain."),
-            _("If it is disabled, you have to choose a server you want to use. Electrum will warn you if your server is lagging.")
-        ])
-        grid.addWidget(self.autoconnect_cb, 4, 0, 1, 3)
-        grid.addWidget(HelpButton(msg), 4, 4)
         self.nodes_list_widget = NodesListWidget(self)
         grid.addWidget(self.nodes_list_widget, 5, 0, 1, 5)
 
@@ -308,14 +314,15 @@ class NetworkChoiceLayout(object):
         self.td = td = TorDetector()
         td.found_proxy.connect(self.suggest_proxy)
         td.start()
+
+        self.fill_in_proxy_settings()
         self.update()
 
-    def check_disable_proxy(self, index = False):
-        if self.config.is_modifiable('proxy'):
-            for w in [self.proxy_host, self.proxy_port, self.proxy_user, self.proxy_password]:
-                w.setEnabled(self.proxy_mode.currentText() != 'NONE')
-        else:
-            for w in [self.proxy_host, self.proxy_port, self.proxy_mode]: w.setEnabled(False)
+    def check_disable_proxy(self, b):
+        if not self.config.is_modifiable('proxy'):
+            b = False
+        for w in [self.proxy_mode, self.proxy_host, self.proxy_port, self.proxy_user, self.proxy_password]:
+            w.setEnabled(b)
 
     def enable_set_server(self):
         if self.config.is_modifiable('server'):
@@ -323,21 +330,18 @@ class NetworkChoiceLayout(object):
             self.server_host.setEnabled(enabled)
             self.server_port.setEnabled(enabled)
             self.servers_list.setEnabled(enabled)
-            self.tabs.setTabEnabled(2, enabled)
         else:
-            for w in [self.autoconnect_cb, self.server_host, self.server_port, self.ssl_cb, self.servers_list]:
+            for w in [self.autoconnect_cb, self.server_host, self.server_port, self.servers_list]:
                 w.setEnabled(False)
 
     def update(self):
         host, port, protocol, proxy_config, auto_connect = self.network.get_parameters()
-        if not proxy_config:
-            proxy_config = { "mode":"none", "host":"localhost", "port":"9050"}
         self.server_host.setText(host)
         self.server_port.setText(port)
-        self.ssl_cb.setChecked(protocol=='s')
         self.autoconnect_cb.setChecked(auto_connect)
 
-        host = self.network.interface.host if self.network.interface else _('None')
+        interface = self.network.interface
+        host = interface.host if interface else _('None')
         self.server_label.setText(host)
 
         self.set_protocol(protocol)
@@ -345,24 +349,17 @@ class NetworkChoiceLayout(object):
         self.servers_list.update(self.servers, self.protocol, self.tor_cb.isChecked())
         self.enable_set_server()
 
-        # proxy tab
-        self.proxy_mode.setCurrentIndex(self.proxy_mode.findText(str(proxy_config.get("mode").upper())))
-        self.proxy_host.setText(proxy_config.get("host"))
-        self.proxy_port.setText(proxy_config.get("port"))
-        self.proxy_user.setText(proxy_config.get("user", ""))
-        self.proxy_password.setText(proxy_config.get("password", ""))
-
         height_str = "%d "%(self.network.get_local_height()) + _('blocks')
         self.height_label.setText(height_str)
         n = len(self.network.get_interfaces())
-        status = _("Connected to %d nodes.")%n if n else _("Not connected")
+        status = _("Connected to {0} nodes.").format(n) if n else _("Not connected")
         self.status_label.setText(status)
         chains = self.network.get_blockchains()
         if len(chains)>1:
             chain = self.network.blockchain()
             checkpoint = chain.get_checkpoint()
             name = chain.get_name()
-            msg = _('Chain split detected at block %d')%checkpoint + '\n'
+            msg = _('Chain split detected at block {0}').format(checkpoint) + '\n'
             msg += (_('You are following branch') if auto_connect else _('Your server is on branch'))+ ' ' + name
             msg += ' (%d %s)' % (chain.get_branch_size(), _('blocks'))
         else:
@@ -370,6 +367,23 @@ class NetworkChoiceLayout(object):
         self.split_label.setText(msg)
         self.nodes_list_widget.update(self.network)
 
+    def fill_in_proxy_settings(self):
+        host, port, protocol, proxy_config, auto_connect = self.network.get_parameters()
+        if not proxy_config:
+            proxy_config = {"mode": "none", "host": "localhost", "port": "9050"}
+
+        b = proxy_config.get('mode') != "none"
+        self.check_disable_proxy(b)
+        if b:
+            self.proxy_cb.setChecked(True)
+            self.proxy_mode.setCurrentIndex(
+                self.proxy_mode.findText(str(proxy_config.get("mode").upper())))
+
+        self.proxy_host.setText(proxy_config.get("host"))
+        self.proxy_port.setText(proxy_config.get("port"))
+        self.proxy_user.setText(proxy_config.get("user", ""))
+        self.proxy_password.setText(proxy_config.get("password", ""))
+
     def layout(self):
         return self.layout_
 
@@ -379,10 +393,10 @@ class NetworkChoiceLayout(object):
 
     def change_protocol(self, use_ssl):
         p = 's' if use_ssl else 't'
-        host = unicode(self.server_host.text())
-        pp = self.servers.get(host, DEFAULT_PORTS)
+        host = self.server_host.text()
+        pp = self.servers.get(host, constants.net.DEFAULT_PORTS)
         if p not in pp.keys():
-            p = pp.keys()[0]
+            p = list(pp.keys())[0]
         port = pp[p]
         self.server_host.setText(host)
         self.server_port.setText(port)
@@ -396,7 +410,7 @@ class NetworkChoiceLayout(object):
     def follow_server(self, server):
         self.network.switch_to_interface(server)
         host, port, protocol, proxy, auto_connect = self.network.get_parameters()
-        host, port, protocol = server.split(':')
+        host, port, protocol = deserialize_server(server)
         self.network.set_parameters(host, port, protocol, proxy, auto_connect)
         self.update()
 
@@ -405,7 +419,7 @@ class NetworkChoiceLayout(object):
             self.change_server(str(x.text(0)), self.protocol)
 
     def change_server(self, host, protocol):
-        pp = self.servers.get(host, DEFAULT_PORTS)
+        pp = self.servers.get(host, constants.net.DEFAULT_PORTS)
         if protocol and protocol not in protocol_letters:
             protocol = None
         if protocol:
@@ -417,11 +431,10 @@ class NetworkChoiceLayout(object):
                 protocol = 's'
                 port = pp.get(protocol)
             else:
-                protocol = pp.keys()[0]
+                protocol = list(pp.keys())[0]
                 port = pp.get(protocol)
         self.server_host.setText(host)
         self.server_port.setText(port)
-        self.ssl_cb.setChecked(protocol=='s')
 
     def accept(self):
         pass
@@ -430,13 +443,12 @@ class NetworkChoiceLayout(object):
         host, port, protocol, proxy, auto_connect = self.network.get_parameters()
         host = str(self.server_host.text())
         port = str(self.server_port.text())
-        protocol = 's' if self.ssl_cb.isChecked() else 't'
         auto_connect = self.autoconnect_cb.isChecked()
         self.network.set_parameters(host, port, protocol, proxy, auto_connect)
 
     def set_proxy(self):
         host, port, protocol, proxy, auto_connect = self.network.get_parameters()
-        if self.proxy_mode.currentText() != 'NONE':
+        if self.proxy_cb.isChecked():
             proxy = { 'mode':str(self.proxy_mode.currentText()).lower(),
                       'host':str(self.proxy_host.text()),
                       'port':str(self.proxy_port.text()),
@@ -444,29 +456,35 @@ class NetworkChoiceLayout(object):
                       'password':str(self.proxy_password.text())}
         else:
             proxy = None
+            self.tor_cb.setChecked(False)
         self.network.set_parameters(host, port, protocol, proxy, auto_connect)
 
     def suggest_proxy(self, found_proxy):
         self.tor_proxy = found_proxy
         self.tor_cb.setText("Use Tor proxy at port " + str(found_proxy[1]))
-        if self.proxy_mode.currentIndex() == 2 \
+        if self.proxy_mode.currentIndex() == self.proxy_mode.findText('SOCKS5') \
             and self.proxy_host.text() == "127.0.0.1" \
                 and self.proxy_port.text() == str(found_proxy[1]):
             self.tor_cb.setChecked(True)
         self.tor_cb.show()
 
     def use_tor_proxy(self, use_it):
-        # 2 = SOCKS5
         if not use_it:
-            self.proxy_mode.setCurrentIndex(0)
-            self.tor_cb.setChecked(False)
+            self.proxy_cb.setChecked(False)
         else:
-            self.proxy_mode.setCurrentIndex(2)
+            socks5_mode_index = self.proxy_mode.findText('SOCKS5')
+            if socks5_mode_index == -1:
+                print_error("[network_dialog] can't find proxy_mode 'SOCKS5'")
+                return
+            self.proxy_mode.setCurrentIndex(socks5_mode_index)
             self.proxy_host.setText("127.0.0.1")
             self.proxy_port.setText(str(self.tor_proxy[1]))
             self.proxy_user.setText("")
             self.proxy_password.setText("")
             self.tor_cb.setChecked(True)
+            self.proxy_cb.setChecked(True)
+        self.check_disable_proxy(use_it)
+        self.set_proxy()
 
     def proxy_settings_changed(self):
         self.tor_cb.setChecked(False)
@@ -489,12 +507,12 @@ class TorDetector(QThread):
     @staticmethod
     def is_tor_port(port):
         try:
-            s = socket._socketobject(socket.AF_INET, socket.SOCK_STREAM)
+            s = (socket._socketobject if hasattr(socket, "_socketobject") else socket.socket)(socket.AF_INET, socket.SOCK_STREAM)
             s.settimeout(0.1)
             s.connect(("127.0.0.1", port))
             # Tor responds uniquely to HTTP-like requests
-            s.send("GET\n")
-            if "Tor is not an HTTP Proxy" in s.recv(1024):
+            s.send(b"GET\n")
+            if b"Tor is not an HTTP Proxy" in s.recv(1024):
                 return True
         except socket.error:
             pass
