@@ -51,7 +51,10 @@ class RequestList(MyTreeWidget):
         if not item.isSelected():
             return
         addr = str(item.text(1))
-        req = self.wallet.receive_requests[addr]
+        req = self.wallet.receive_requests.get(addr)
+        if req is None:
+            self.update()
+            return
         expires = age(req['time'] + req['exp']) if req.get('exp') else _('Never')
         amount = req['amount']
         message = self.wallet.labels.get(addr, '')
@@ -110,7 +113,10 @@ class RequestList(MyTreeWidget):
         if not item:
             return
         addr = str(item.text(1))
-        req = self.wallet.receive_requests[addr]
+        req = self.wallet.receive_requests.get(addr)
+        if req is None:
+            self.update()
+            return
         column = self.currentColumn()
         column_title = self.headerItem().text(column)
         column_data = item.text(column)
