@@ -10,6 +10,7 @@ from electrum.bitcoin import serialize_xpub
 class GuiMixin(object):
     # Requires: self.proto, self.device
 
+    # ref: https://github.com/trezor/trezor-common/blob/44dfb07cfaafffada4b2ce0d15ba1d90d17cf35e/protob/types.proto#L89
     messages = {
         3: _("Confirm the transaction output on your {} device"),
         4: _("Confirm internal entropy on your {} device to begin"),
@@ -19,6 +20,7 @@ class GuiMixin(object):
         8: _("Confirm the total amount spent and the transaction fee on your "
              "{} device"),
         10: _("Confirm wallet address on your {} device"),
+        14: _("Choose on your {} device where to enter your passphrase"),
         'default': _("Check your {} device to continue"),
     }
 
@@ -84,6 +86,15 @@ class GuiMixin(object):
         return self.proto.PassphraseStateAck()
 
     def callback_WordRequest(self, msg):
+        if (msg.type is not None
+            and msg.type in (self.types.WordRequestType.Matrix9,
+                             self.types.WordRequestType.Matrix6)):
+            num = 9 if msg.type == self.types.WordRequestType.Matrix9 else 6
+            char = self.handler.get_matrix(num)
+            if char == 'x':
+                return self.proto.Cancel()
+            return self.proto.WordAck(word=char)
+
         self.step += 1
         msg = _("Step {}/24.  Enter seed word as explained on "
                 "your {}:").format(self.step, self.device)
@@ -91,12 +102,6 @@ class GuiMixin(object):
         # Unfortunately the device can't handle self.proto.Cancel()
         return self.proto.WordAck(word=word)
 
-    def callback_CharacterRequest(self, msg):
-        char_info = self.handler.get_char(msg)
-        if not char_info:
-            return self.proto.Cancel()
-        return self.proto.CharacterAck(**char_info)
-
 
 class TrezorClientBase(GuiMixin, PrintError):
 
@@ -230,6 +235,10 @@ class TrezorClientBase(GuiMixin, PrintError):
     def atleast_version(self, major, minor=0, patch=0):
         return self.firmware_version() >= (major, minor, patch)
 
+    def get_trezor_model(self):
+        """Returns '1' for Trezor One, 'T' for Trezor T."""
+        return self.features.model
+
     @staticmethod
     def wrapper(func):
         '''Wrap methods to clear any message box they opened.'''
