@@ -27,11 +27,12 @@ import socket
 import os
 import requests
 import json
+import base64
 from urllib.parse import urljoin
 from urllib.parse import quote
 
 import electrum
-from electrum import bitcoin
+from electrum import bitcoin, ecc
 from electrum import constants
 from electrum import keystore
 from electrum.bitcoin import *
@@ -75,6 +76,18 @@ DISCLAIMER = [
       "To be safe from malware, you may want to do this on an offline "
       "computer, and move your wallet later to an online computer."),
 ]
+
+KIVY_DISCLAIMER = [
+    _("Two-factor authentication is a service provided by TrustedCoin. "
+      "To use it, you must have a separate device with Google Authenticator."),
+    _("This service uses a multi-signature wallet, where you own 2 of 3 keys.  "
+      "The third key is stored on a remote server that signs transactions on "
+      "your behalf. A small fee will be charged on each transaction that uses the "
+      "remote server."),
+    _("Note that your coins are not locked in this service.  You may withdraw "
+      "your funds at any time and at no cost, without the remote server, by "
+      "using the 'restore wallet' option with your wallet seed."),
+]
 RESTORE_MSG = _("Enter the seed for your 2-factor wallet:")
 
 class TrustedCoinException(Exception):
@@ -82,6 +95,11 @@ class TrustedCoinException(Exception):
         Exception.__init__(self, message)
         self.status_code = status_code
 
+
+class ErrorConnectingServer(Exception):
+    pass
+
+
 class TrustedCoinCosignerClient(object):
     def __init__(self, user_agent=None, base_url='https://api.trustedcoin.com/2/'):
         self.base_url = base_url
@@ -100,7 +118,10 @@ class TrustedCoinCosignerClient(object):
         url = urljoin(self.base_url, relative_url)
         if self.debug:
             print('%s %s %s' % (method, url, data))
-        response = requests.request(method, url, **kwargs)
+        try:
+            response = requests.request(method, url, **kwargs)
+        except Exception as e:
+            raise ErrorConnectingServer(e)
         if self.debug:
             print(response.text)
         if response.status_code != 200:
@@ -236,7 +257,8 @@ class Wallet_2fa(Multisig_Wallet):
             return 0
         n = self.num_prepay(config)
         price = int(self.price_per_tx[n])
-        assert price <= 100000 * n
+        if price > 100000 * n:
+            raise Exception('too high trustedcoin fee ({} for {} txns)'.format(price, n))
         return price
 
     def make_unsigned_transaction(self, coins, outputs, config, fixed_fee=None,
@@ -260,23 +282,22 @@ class Wallet_2fa(Multisig_Wallet):
             tx = mk_tx(outputs)
         return tx
 
-    def sign_transaction(self, tx, password):
-        Multisig_Wallet.sign_transaction(self, tx, password)
-        if tx.is_complete():
-            return
-        if not self.auth_code:
+    def on_otp(self, tx, otp):
+        if not otp:
             self.print_error("sign_transaction: no auth code")
             return
+        otp = int(otp)
         long_user_id, short_id = self.get_user_id()
-        tx_dict = tx.as_dict()
-        raw_tx = tx_dict["hex"]
-        r = server.sign(short_id, raw_tx, self.auth_code)
+        raw_tx = tx.serialize_to_network()
+        r = server.sign(short_id, raw_tx, otp)
         if r:
             raw_tx = r.get('transaction')
             tx.update(raw_tx)
         self.print_error("twofactor: is complete", tx.is_complete())
         # reset billing_info
         self.billing_info = None
+        self.plugin.start_request_thread(self)
+
 
 
 # Utility functions
@@ -305,6 +326,7 @@ def make_billing_address(wallet, num):
 
 class TrustedCoinPlugin(BasePlugin):
     wallet_class = Wallet_2fa
+    disclaimer_msg = DISCLAIMER
 
     def __init__(self, parent, config, name):
         BasePlugin.__init__(self, parent, config, name)
@@ -324,29 +346,60 @@ class TrustedCoinPlugin(BasePlugin):
     def can_user_disable(self):
         return False
 
+    @hook
+    def tc_sign_wrapper(self, wallet, tx, on_success, on_failure):
+        if not isinstance(wallet, self.wallet_class):
+            return
+        if tx.is_complete():
+            return
+        if wallet.can_sign_without_server():
+            return
+        if not wallet.keystores['x3/'].get_tx_derivations(tx):
+            self.print_error("twofactor: xpub3 not needed")
+            return
+        def wrapper(tx):
+            self.prompt_user_for_otp(wallet, tx, on_success, on_failure)
+        return wrapper
+
     @hook
     def get_tx_extra_fee(self, wallet, tx):
         if type(wallet) != Wallet_2fa:
             return
         if wallet.billing_info is None:
-            assert wallet.can_sign_without_server()
+            if not wallet.can_sign_without_server():
+                self.start_request_thread(wallet)
+                raise Exception('missing trustedcoin billing info')
             return None
         address = wallet.billing_info['billing_address']
         for _type, addr, amount in tx.outputs():
             if _type == TYPE_ADDRESS and addr == address:
                 return address, amount
 
+    def finish_requesting(func):
+        def f(self, *args, **kwargs):
+            try:
+                return func(self, *args, **kwargs)
+            finally:
+                self.requesting = False
+        return f
+
+    @finish_requesting
     def request_billing_info(self, wallet):
         if wallet.can_sign_without_server():
             return
         self.print_error("request billing info")
-        billing_info = server.get(wallet.get_user_id()[1])
+        try:
+            billing_info = server.get(wallet.get_user_id()[1])
+        except ErrorConnectingServer as e:
+            self.print_error('cannot connect to TrustedCoin server: {}'.format(e))
+            return
         billing_address = make_billing_address(wallet, billing_info['billing_index'])
-        assert billing_address == billing_info['billing_address']
+        if billing_address != billing_info['billing_address']:
+            raise Exception('unexpected trustedcoin billing address: expected {}, received {}'
+                            .format(billing_address, billing_info['billing_address']))
         wallet.billing_info = billing_info
         wallet.price_per_tx = dict(billing_info['price_per_tx'])
-        wallet.price_per_tx.pop(1)
-        self.requesting = False
+        wallet.price_per_tx.pop(1, None)
         return True
 
     def start_request_thread(self, wallet):
@@ -368,7 +421,7 @@ class TrustedCoinPlugin(BasePlugin):
     def show_disclaimer(self, wizard):
         wizard.set_icon(':icons/trustedcoin-wizard.png')
         wizard.stack = []
-        wizard.confirm_dialog(title='Disclaimer', message='\n\n'.join(DISCLAIMER), run_next = lambda x: wizard.run('choose_seed'))
+        wizard.confirm_dialog(title='Disclaimer', message='\n\n'.join(self.disclaimer_msg), run_next = lambda x: wizard.run('choose_seed'))
 
     def choose_seed(self, wizard):
         title = _('Create or restore')
@@ -398,8 +451,12 @@ class TrustedCoinPlugin(BasePlugin):
         words = seed.split()
         n = len(words)
         # old version use long seed phrases
-        if n >= 24:
-            assert passphrase == ''
+        if n >= 20:
+            # note: pre-2.7 2fa seeds were typically 24-25 words, however they
+            # could probabilistically be arbitrarily shorter due to a bug. (see #3611)
+            # the probability of it being < 20 words is about 2^(-(256+12-19*11)) = 2^(-59)
+            if passphrase != '':
+                raise Exception('old 2fa seed cannot have passphrase')
             xprv1, xpub1 = self.get_xkeys(' '.join(words[0:12]), '', "m/")
             xprv2, xpub2 = self.get_xkeys(' '.join(words[12:]), '', "m/")
         elif n==12:
@@ -424,18 +481,7 @@ class TrustedCoinPlugin(BasePlugin):
         wizard.storage.put('x1/', k1.dump())
         wizard.storage.put('x2/', k2.dump())
         wizard.storage.write()
-        msg = [
-            _("Your wallet file is: {}.").format(os.path.abspath(wizard.storage.path)),
-            _("You need to be online in order to complete the creation of "
-              "your wallet.  If you generated your seed on an offline "
-              'computer, click on "{}" to close this window, move your '
-              "wallet file to an online computer, and reopen it with "
-              "Electrum.").format(_('Cancel')),
-            _('If you are online, click on "{}" to continue.').format(_('Next'))
-        ]
-        msg = '\n\n'.join(msg)
-        wizard.stack = []
-        wizard.confirm_dialog(title='', message=msg, run_next = lambda x: wizard.run('create_remote_key'))
+        self.go_online_dialog(wizard)
 
     def restore_wallet(self, wizard):
         wizard.opt_bip39 = False
@@ -490,8 +536,8 @@ class TrustedCoinPlugin(BasePlugin):
         wizard.wallet = Wallet_2fa(storage)
         wizard.create_addresses()
 
-    def create_remote_key(self, wizard):
-        email = self.accept_terms_of_use(wizard)
+
+    def create_remote_key(self, email, wizard):
         xpub1 = wizard.storage.get('x1/')['xpub']
         xpub2 = wizard.storage.get('x2/')['xpub']
         # Generate third key deterministically.
@@ -500,8 +546,9 @@ class TrustedCoinPlugin(BasePlugin):
         # secret must be sent by the server
         try:
             r = server.create(xpub1, xpub2, email)
-        except socket.error:
+        except (socket.error, ErrorConnectingServer):
             wizard.show_message('Server not reachable, aborting')
+            wizard.terminate()
             return
         except TrustedCoinException as e:
             if e.status_code == 409:
@@ -518,16 +565,17 @@ class TrustedCoinPlugin(BasePlugin):
                 return
             _xpub3 = r['xpubkey_cosigner']
             _id = r['id']
-            try:
-                assert _id == short_id, ("user id error", _id, short_id)
-                assert xpub3 == _xpub3, ("xpub3 error", xpub3, _xpub3)
-            except Exception as e:
-                wizard.show_message(str(e))
+            if short_id != _id:
+                wizard.show_message("unexpected trustedcoin short_id: expected {}, received {}"
+                                    .format(short_id, _id))
                 return
-        self.check_otp(wizard, short_id, otp_secret, xpub3)
+            if xpub3 != _xpub3:
+                wizard.show_message("unexpected trustedcoin xpub3: expected {}, received {}"
+                                    .format(xpub3, _xpub3))
+                return
+        self.request_otp_dialog(wizard, short_id, otp_secret, xpub3)
 
-    def check_otp(self, wizard, short_id, otp_secret, xpub3):
-        otp, reset = self.request_otp_dialog(wizard, short_id, otp_secret)
+    def check_otp(self, wizard, short_id, otp_secret, xpub3, otp, reset):
         if otp:
             self.do_auth(wizard, short_id, otp, xpub3)
         elif reset:
@@ -543,22 +591,29 @@ class TrustedCoinPlugin(BasePlugin):
     def do_auth(self, wizard, short_id, otp, xpub3):
         try:
             server.auth(short_id, otp)
-        except:
-            wizard.show_message(_('Incorrect password'))
-            return
-        k3 = keystore.from_xpub(xpub3)
-        wizard.storage.put('x3/', k3.dump())
-        wizard.storage.put('use_trustedcoin', True)
-        wizard.storage.write()
-        wizard.wallet = Wallet_2fa(wizard.storage)
-        wizard.run('create_addresses')
+        except TrustedCoinException as e:
+            if e.status_code == 400:  # invalid OTP
+                wizard.show_message(_('Invalid one-time password.'))
+                # ask again for otp
+                self.request_otp_dialog(wizard, short_id, None, xpub3)
+            else:
+                wizard.show_message(str(e))
+                wizard.terminate()
+        except Exception as e:
+            wizard.show_message(str(e))
+            wizard.terminate()
+        else:
+            k3 = keystore.from_xpub(xpub3)
+            wizard.storage.put('x3/', k3.dump())
+            wizard.storage.put('use_trustedcoin', True)
+            wizard.storage.write()
+            wizard.wallet = Wallet_2fa(wizard.storage)
+            wizard.run('create_addresses')
 
     def on_reset_auth(self, wizard, short_id, seed, passphrase, xpub3):
         xprv1, xpub1, xprv2, xpub2 = self.xkeys_from_seed(seed, passphrase)
-        try:
-            assert xpub1 == wizard.storage.get('x1/')['xpub']
-            assert xpub2 == wizard.storage.get('x2/')['xpub']
-        except:
+        if (wizard.storage.get('x1/')['xpub'] != xpub1 or
+                wizard.storage.get('x2/')['xpub'] != xpub2):
             wizard.show_message(_('Incorrect seed'))
             return
         r = server.get_challenge(short_id)
@@ -567,7 +622,7 @@ class TrustedCoinPlugin(BasePlugin):
         def f(xprv):
             _, _, _, _, c, k = deserialize_xprv(xprv)
             pk = bip32_private_key([0, 0], k, c)
-            key = regenerate_key(pk)
+            key = ecc.ECPrivkey(pk)
             sig = key.sign_message(message, True)
             return base64.b64encode(sig).decode()
 
@@ -577,7 +632,7 @@ class TrustedCoinPlugin(BasePlugin):
         if not new_secret:
             wizard.show_message(_('Request rejected by server'))
             return
-        self.check_otp(wizard, short_id, new_secret, xpub3)
+        self.request_otp_dialog(wizard, short_id, new_secret, xpub3)
 
     @hook
     def get_action(self, storage):
@@ -588,4 +643,4 @@ class TrustedCoinPlugin(BasePlugin):
         if not storage.get('x2/'):
             return self, 'show_disclaimer'
         if not storage.get('x3/'):
-            return self, 'create_remote_key'
+            return self, 'accept_terms_of_use'
